import datetime as dt

import asyncio
import json

from typing import List, Dict, Optional, Union, TYPE_CHECKING, TypedDict, Any

from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.app import App

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


class Activity(Resource):
    """
    A class that contains all the information necessary to create and run an activity.
    """

    _module_logger = logger
    resource_type = 'activity'
    alias_field = 'name'
    plural = 'activities'

    class ActivityParams(TypedDict):
        name: Optional[str]
        settings: dict
        activityTemplateWithMode: Optional[dict]
        universeApiKeyId: Optional[str]

    class Run(Resource):
        """
        Run of an activity, which is a single execution of the activity.
        """

        _module_logger = logger
        resource_type = 'run'
        plural = 'runs'

        class Log(Resource):
            """
            Log of an activity run.
            """

            _module_logger = logger
            resource_type = 'log'
            plural = 'logs'

            def __init__(self, parent: 'Activity.Run', uuid: Optional[str] = None, db_resource: Optional[Dict] = None):

                params = dict(
                    dateTime=str(dt.datetime.now().isoformat())[:-3] + "Z",
                    message='message',
                    severity='INFO',
                    tags={},
                )

                super().__init__(parent=parent, uuid=uuid, db_resource=db_resource,
                                 check_params_before_creation=['message', 'severity'],
                                 params_to_serialize=['tags'], params=params)

        def __init__(self, parent: 'Activity', uuid: Optional[str] = None, db_resource: Optional[Dict] = None):

            params = dict(
                settings={},
            )

            super().__init__(parent=parent, uuid=uuid, db_resource=db_resource, children=[Activity.Run.Log],
                             params_to_serialize=['settings'], params=params)

        async def create_log(self, message: str, severity: str, tags: Dict[str, str]) -> 'Activity.Run.Log':
            """
            Creates a log for the activity run.
            :param message: The message of the log.
            :param severity: The severity of the log.
            :param tags: The tags of the log.
            :return: The created log.
            """
            if tags and (not isinstance(tags, Dict) or any([not isinstance(key, str) or not isinstance(value, str)
                                                            for key, value in tags.items()])):
                log_error(logger, 'The tags must be a dictionary of strings.', ValueError)

            return await self._base_resource.create_child(Activity.Run.Log,
                                                          message=message, severity=severity, tags=tags)

        async def get_log(self, uuid: str) -> Optional['Activity.Run.Log']:
            """
            Gets a log of the activity run.
            :param uuid: The uuid of the log.
            :return: The log.
            """
            return await self._base_resource.get_child(Activity.Run.Log, uuid)

        async def get_logs(self) -> List['Activity.Run.Log']:
            """
            Gets all the logs of the activity run.
            :return: The logs.
            """
            logs = await self._base_resource.get_children(Activity.Run.Log)
            return sorted(logs, key=lambda log: log['dateTime'])

        async def trigger_webhook(self):
            """
            Triggers the webhook of the activity in a run.
            """
            endpoint = self._base_resource.base_url + f'run/{self._base_resource.id}/triggerWebhook'

            response = await(
                self._base_resource.api_client.query_element(
                    method='POST', endpoint=endpoint,
                )
            )

            return response['STATUS']

        def cascade_to_dict(self) -> Dict[str, Any]:
            """
            Returns the run as a dictionary.
            :return: The run as a dictionary.
            """
            run_dict = super().cascade_to_dict()
            run_dict['logs'] = sorted(run_dict['logs'], key=lambda log: log['dateTime'])
            return run_dict

    @staticmethod
    def _runs_ordering(run: 'Activity.Run') -> dt.datetime:
        """ Returns the datetime of the last log of the run. Used for sorting runs.
        :param run: The run.
        :return: The datetime of the last log of the run.
        """
        logs = run._base_resource.children[Activity.Run.Log]
        if len(logs) > 0:
            return max([
                dt.datetime.strptime(log['dateTime'], "%Y-%m-%dT%H:%M:%S.%fZ")
                if log['dateTime'] else dt.datetime.min
                for log_id, log in logs
            ])
        else:
            return dt.datetime.min

    @staticmethod
    async def sort_runs_by_log_time(runs: List['Activity.Run']) -> List['Activity.Run']:
        """ Returns the runs ordered by the datetime of the last log of the run.
        :param runs: The runs.
        :return: The runs ordered by the datetime of the last log of the run.
        """
        await asyncio.gather(*[run.get_logs() for run in runs])
        return sorted(runs, key=Activity._runs_ordering)

    def __init__(self, parent: 'App', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[Dict] = None):

        params: Activity.ActivityParams = dict(
            name=alias,
            settings={},
            activityTemplateWithMode=None,
            universeApiKeyId=None,
        )

        super().__init__(
            parent=parent, uuid=uuid, db_resource=db_resource, params=params,
            children=[Activity.Run], check_params_before_creation=['name'],
            params_to_serialize=['settings'],

        )

    async def delete(self):
        """ Deletes the activity. """
        return await self._base_resource.delete()

    async def update(self):
        """ Updates the activity. """
        return await self._base_resource.update()

    # Webhook methods
    async def create_webhook(self, url: str, method: str, headers: Dict[str, str]):
        """ Creates a webhook for the activity.
        :param url: The url of the webhook.
        :param method: The method of the webhook.
        :param headers: The headers of the webhook.
        """
        if not headers:
            headers = {}
        if not isinstance(headers, Dict) or any([not isinstance(key, str) or not isinstance(value, str)
                                                for key, value in headers.items()]):
            log_error(logger, 'The headers must be a dictionary of strings.', ValueError)

        endpoint = self._base_resource.base_url + f'activity/{self._base_resource.id}/webhook'
        params = dict(
            url=url,
            method=method,
            headers=json.dumps(headers),
        )
        await self._base_resource.api_client.query_element(
            method='POST', endpoint=endpoint,
            **{'body_params': params}
        )

    # Run methods
    async def get_run(self, uuid: str) -> Optional['Activity.Run']:
        """ Gets a run of the activity.
        :param uuid: The uuid of the run.
        :return: The run.
        """
        return await self._base_resource.get_child(Activity.Run, uuid)

    async def get_runs(self, how_many_runs: Optional[int] = None) -> List['Activity.Run']:
        """ Gets the last runs of the activity.
        :param how_many_runs: The number of runs to get.
        :return: The runs.
        """
        runs = await Activity.sort_runs_by_log_time(await self._base_resource.get_children(Activity.Run))
        return runs[-how_many_runs:] if how_many_runs is not None else runs

    async def create_run(self, settings: Optional[Union[Dict, str]] = None) -> 'Activity.Run':
        """ Creates a run of the activity.
        :param settings: The settings of the run or id of the run to copy the settings from.
        :return: The run.
        """
        if isinstance(settings, str):
            settings = (await self.get_run(settings))['settings']
            assert isinstance(settings, dict)

        return await self._base_resource.create_child(Activity.Run, settings=settings)



from ...report import Report


class Indicator(Report):
    report_type = 'INDICATOR'

    possible_values = dict(
        colors=['success', 'warning', 'error', 'neutral', 'caution'],
        align=['left', 'center', 'right'],
        variant=['default', 'outlined', 'contained', 'topColor'],
    )

    default_properties = dict(
        **Report.default_properties,
        title=None,
        value=None,
        description=None,
        targetPath=None,
        color='neutral',
        align='right',
        variant='default',
        info=None,
        icon=None,
        bigIcon=None,
        hideIcons=False,
        backgroundImage=None,
    )



from enum import Enum
import inspect
import unicodedata
from typing import Optional
from shimoku.execution_logger import log_error
import sys

import logging
logger = logging.getLogger(__name__)

IN_BROWSER_PYODIDE = "_pyodide_core" in sys.modules
ACTIONS_TEST = False


class EventType(Enum):
    NO_EVENT = "NO_EVENT"
    EVENT = "EVENT"
    REPORT_UPDATED = "REPORT_UPDATED"
    REPORT_CREATED = "REPORT_CREATED"
    REPORT_DELETED = "REPORT_DELETED"
    APP_UPDATED = "APP_UPDATED"
    APP_CREATED = "APP_CREATED"
    APP_DELETED = "APP_DELETED"
    DASHBOARD_UPDATED = "DASHBOARD_UPDATED"
    DASHBOARD_CREATED = "DASHBOARD_CREATED"
    DASHBOARD_DELETED = "DASHBOARD_DELETED"
    BUSINESS_CONTENTS_UPDATED = "BUSINESS_CONTENTS_UPDATED"


def get_arg_names(func):
    signature = inspect.signature(func)
    return [k for k, v in signature.parameters.items()]


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default if v.default is not inspect.Parameter.empty else None
        for k, v in signature.parameters.items()
    }


def get_args_with_defaults(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def create_function_name(name: Optional[str]) -> str:
    """ Create a valid function name from a string
    :param name: string to create function name from
    :return: valid function name
    """

    def check_correct_character(character: str) -> bool:
        """ Check if a character is valid for a function name
        :param character: character to check
        :return: True if character is valid, False otherwise
        """
        return character.isalnum() or character in ['_', '-', ' ']

    if name is None:
        return 'no_path'

    # Normalize to ASCII
    normalized_name = unicodedata.normalize('NFKD', name).encode('ascii', 'ignore').decode('ascii')

    # Change Uppercase to '_' + lowercase if previous character is in abecedary
    normalized_name = ''.join(['_' + c.lower() if c.isupper() and i > 0 and normalized_name[i - 1].isalpha() else c
                               for i, c in enumerate(normalized_name) if check_correct_character(c)])

    return create_normalized_name(normalized_name).replace('-', '_')


def create_normalized_name(name: str) -> str:
    """Having a name create a normalizedName

    Example
    ----------------------
    # "name": "   Test Borrar_grafico    "
    # "normalizedName": "test-borrar-grafico"
    """
    if any([c not in 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._- &$'
            for c in name.encode('ascii', 'ignore').decode()]):
        log_error(logger,
                  f'You can only use letters, numbers, spaces, "-" and "_" in your name | '
                  f'you introduced {name}',
                  ValueError
                  )
    # remove empty spaces at the beginning and end
    name: str = name.strip()
    # replace "_" for www protocol it is not good
    name = name.replace('_', '-')

    return '-'.join(name.split(' ')).lower()


def clean_menu_path(menu_path: str) -> tuple[str, str]:
    """
    Break the menu path in the apptype or app normalizedName
    and the path normalizedName if any
    :param menu_path: the menu path
    """
    # remove empty spaces
    menu_path: str = menu_path.strip()
    # replace "_" for www protocol it is not good
    menu_path = menu_path.replace('_', '-')

    try:
        assert len(menu_path.split('/')) <= 2  # we allow only one level of path
    except AssertionError:
        raise ValueError(
            f'We only allow one subpath in your request | '
            f'you introduced {menu_path} it should be maximum '
            f'{"/".join(menu_path.split("/")[:1])}'
        )

    # Split AppType or App Normalized Name
    normalized_name: str = menu_path.split('/')[0]
    name: str = (
        ' '.join(normalized_name.split('-'))
    )

    try:
        path_normalized_name: str = menu_path.split('/')[1]
        path_name: str = (
            ' '.join(path_normalized_name.split('-'))
        )
    except IndexError:
        path_name = None

    return name, path_name



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class ComponentsLayerHeader:
    """
    This class is used to interact with the API at the component level.
    """

    def delete_component(
        self,
        uuid: str,
    ):
        """
        :param uuid: uuid of the component
        """
        pass

    def get_component(
        self,
        uuid: str,
    ) -> dict:
        """
        :param uuid: uuid of the component
        :return: component
        """
        pass

    def get_component_data_set_links(
        self,
        uuid: str,
    ) -> list:
        """
        :param uuid: uuid of the component
        :return: list of data set links
        """
        pass

    def get_components_in_sub_path(
        self,
        path: str,
    ) -> list:
        """
        :param path: path to the components
        :return: list of components
        """
        pass



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class BoardsLayerHeader:
    """
    This class is used to interact with the API at the board level.
    """

    def add_menu_path_in_board(
        self,
        menu_path_name: Optional[str] = None,
        menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Add an app in a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def create_board(
        self,
        name: str,
        order: Optional[int] = None,
        is_public: bool = False,
        is_disabled: bool = False,
    ) -> dict:
        """
        Create a board
        :param name: name of the board
        :param order: order of the board
        :param is_public: is the board public
        :param is_disabled: is the board disabled
        :return: board metadata
        """
        pass

    def create_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        resource: Optional[str] = None,
        role_name: Optional[str] = None,
        permission: Optional[str] = None,
        target: Optional[str] = None,
    ) -> dict:
        """
    Create a new role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param resource: resource of the role
    :param role_name: name of the role
    :param permission: permission of the role
    :param target: target of the role
    """
        pass

    def delete_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """
        Delete a board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def delete_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_id: Optional[str] = None,
        role_name: Optional[str] = None,
    ):
        """
    Delete the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_id: id of the role
    :param role_name: name of the role
    """
        pass

    def force_delete_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Force delete a board, this will delete the links between the board and the boards first, so that
        the board can always be deleted, but the apps will not be deleted
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def get_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Optional[dict]:
        """
        Get the board metadata
        :param name: name of the board
        :param uuid: UUID of the board
        :return: board metadata
        """
        pass

    def get_board_menu_path_ids(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the list of the ids of the boards in the board
        :param name: name of the board
        :param uuid: UUID of the board
        :return: list of the ids of the boards in the board
        """
        pass

    def get_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_name: Optional[str] = None,
    ) -> list:
        """
    Get the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_name: name of the role
    """
        pass

    def get_roles(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
    Get the roles at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    """
        pass

    def group_menu_paths(
        self,
        menu_path_names: Union[list[str], None, str] = None,
        menu_path_ids: Optional[list[str]] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Add multiple boards in a board, if the board does not exist create it
        :param menu_path_names: list of board names
        :param menu_path_ids: list of board UUIDs
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def is_menu_path_in_board(
        self,
        menu_path_name: Optional[str] = None,
        menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """
        Check if an app is in a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        :return: True if the app is in the board, False otherwise
        """
        pass

    def remove_all_menu_paths_from_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete all boards links of a board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def remove_menu_path_from_board(
        self,
        menu_path_name: Optional[str] = None,
        menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Remove an app from a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def update_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        order: Optional[int] = None,
        is_public: Optional[bool] = None,
        is_disabled: Optional[bool] = None,
    ) -> bool:
        """
        Update a board
        :param new_name: name of the board
        :param uuid: UUID of the board
        :param name: new name of the board
        :param order: new order of the board
        :param is_public: new public permission of the board
        :param is_disabled: new is_disabled of the board
        """
        pass



from typing import Optional
from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import ResourceGetter, InitOptions
from shimoku.cli.utils import choose_from_menu
from shimoku.execution_logger import configure_logging

import logging
logger = logging.getLogger(__name__)


def add_update_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the update parser to a parser
    :param parser: Parser to add the update parser to
    :return: Update parser
    """
    params = {
        'name': 'update',
        'description': 'Commands to update existing resources'
    }
    if parser:
        update_parser = parser.add_command(**params)
    else:
        update_parser = CLIParser(**params)

    common_args = [
        CLIFuncParam(name='local-port', arg_type=int, arg_help='Local port to use', mandatory=False),
        CLIFuncParam(name='environment', arg_type=str, arg_help='Environment to use', mandatory=False),
        CLIFuncParam(name='access-token', arg_type=str, arg_help='Access token to use', mandatory=False),
        CLIFuncParam(name='universe-id', arg_type=str, arg_help='Universe ID to use', mandatory=False),
        CLIFuncParam(name='workspace-id', arg_type=str, arg_help='Workspace ID to use', mandatory=False),
    ]

    module_functions = [
        workspace,
        menu_path,
        menu_order,
        board,
        boards_order
    ]

    for func in module_functions:
        update_parser.decor_add_command(common_args=common_args)(func)

    return update_parser


if __name__ == "__main__":
    add_update_parser().parse_args()


async def workspace(
    new_name: str = CLIFuncParam(mandatory=False),
    reset_theme: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ Update the name or reset the theme of a workspace
    :param new_name: New name for the workspace
    :param reset_theme: Flag to reset the theme of the workspace
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business_id = resource_getter.init_opts.workspace_id
    if not reset_theme:
        await businesses_layer.update_workspace(business_id, new_name=new_name)
    else:
        await businesses_layer.update_workspace(business_id, new_name=new_name, theme={})


async def menu_order(
    change_sub_paths: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ Update the order of menu paths
    :param change_sub_paths: Flag to change the order of the sub paths
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()

    apps_layer = await resource_getter.get_apps_layer()
    app_names = [app['name'] for app in await business.get_apps()] + ['']

    apps_order = []
    print('Choose the order of the menu paths, press enter to finish')
    previous_logging_level = logging.root.level
    if previous_logging_level == logging.INFO:
        configure_logging('WARNING')
    while True:
        next_app = choose_from_menu(app_names, 'Next menu path: ')
        if next_app:
            if change_sub_paths:
                sub_paths_order = []
                sub_paths = (await apps_layer.get_menu_path_sub_paths(name=next_app)) + ['']
                print(f'Choose the order of the sub paths in the menu path({next_app}), press enter to finish')
                while True:
                    next_sub_path = choose_from_menu(sub_paths, 'Next sub path: ')
                    if next_sub_path:
                        sub_paths_order.append(next_sub_path)
                    else:
                        break
                    sub_paths.remove(next_sub_path)
                    if len(sub_paths) == 1:
                        break
                apps_order.append((next_app, sub_paths_order))
            else:
                apps_order.append(next_app)
        else:
            break
        app_names.remove(next_app)
        if len(app_names) == 1:
            break
    if previous_logging_level == logging.INFO:
        configure_logging('INFO')

    await businesses_layer.change_menu_order(
        uuid=resource_getter.init_opts.workspace_id, menu_order=apps_order
    )


async def boards_order(**kwargs):
    """ Update the order of the boards in a workspace """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()

    dashboard_names = [d['name'] for d in await business.get_dashboards()] + ['']
    dashboards_order = []
    print('Choose the order of the dashboards in the workspace, press enter to finish')
    while True:
        next_dashboard = choose_from_menu(dashboard_names, 'Next board: ')
        if next_dashboard:
            dashboards_order.append(next_dashboard)
        else:
            break
        dashboard_names.remove(next_dashboard)
        if len(dashboard_names) == 1:
            break
    await businesses_layer.change_boards_order(
        uuid=resource_getter.init_opts.workspace_id, boards=dashboards_order
    )


async def board(
    is_public: Optional[str],
    is_disabled: Optional[str],
    board: str = CLIFuncParam(prompt=True),
    new_name: str = CLIFuncParam(mandatory=False),
    order: int = CLIFuncParam(mandatory=False),
    **kwargs
):
    """
    Update the properties of a board
    :param is_public: (true/false) Set the board as public
    :param is_disabled: (true/false) Set the board as disabled
    :param board: Name or id of the board
    :param new_name: New name for the board
    :param order: New order for the board
    """
    if is_public is not None:
        is_public = is_public.lower() == 'true'
    if is_disabled is not None:
        is_disabled = is_disabled.lower() == 'true'
    resource_getter = ResourceGetter(InitOptions(board=board, **kwargs))
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard = await resource_getter.get_dashboard()
    await dashboards_layer.update_board(
        uuid=dashboard['id'],
        new_name=new_name,
        order=order,
        is_public=is_public,
        is_disabled=is_disabled
    )


async def menu_path(
    hide_title: Optional[str],
    hide_path: Optional[str],
    show_breadcrumb: Optional[str],
    show_history_navigation: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    new_name: str = CLIFuncParam(mandatory=False),
    order: int = CLIFuncParam(mandatory=False),
    **kwargs
):
    """
    Update the properties of a menu path
    :param menu_path: Name or id of the menu path
    :param new_name: New name for the menu path
    :param order: New order for the menu path
    :param hide_title: (true/false) Hide the title of the menu path
    :param hide_path: (true/false) Hide the path of the menu path
    :param show_breadcrumb: (true/false) Show the breadcrumb of the menu path
    :param show_history_navigation: (true/false) Show the history navigation of the menu path
    """
    if hide_title is not None:
        hide_title = hide_title.lower() == 'true'
    if hide_path is not None:
        hide_path = hide_path.lower() == 'true'
    if show_breadcrumb is not None:
        show_breadcrumb = show_breadcrumb.lower() == 'true'
    if show_history_navigation is not None:
        show_history_navigation = show_history_navigation.lower() == 'true'
    resource_getter = ResourceGetter(InitOptions(menu_path=menu_path, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()

    await apps_layer.update_menu_path(
        uuid=app['id'],
        new_name=new_name,
        order=order,
        hide_title=hide_title,
        hide_path=hide_path,
        show_breadcrumb=show_breadcrumb,
        show_history_navigation=show_history_navigation
    )




# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class ActivitiesLayerHeader:
    """
    This class is used to interact with the activities that are available in a menu path, through the API.
    """

    def create_activity(
        self,
        name: str,
        settings: Optional[dict] = None,
        template_id: Optional[str] = None,
        template_name_version: Optional[tuple[str, str]] = None,
        template_mode: str = "LIGHT",
        universe_api_key: str = "",
    ) -> dict:
        """
        Create an activity by its name and app id
        :param name: the name of the activity
        :param settings: the settings of the activity
        :param template_id: the template id of the activity
        :param template_name_version: the template name and version of the activity
        :param template_mode: the template mode of the activity
        :param universe_api_key: the universe api key of the activity
        :return: the dictionary representation of the activity
        """
        pass

    def create_run(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        settings: Union[dict, str] = None,
    ) -> dict:
        """
        Create a run for an activity by its name
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param settings: the settings of the run, or the id of the run to clone settings from
        :return: the run created as a dictionary
        """
        pass

    def create_run_log(
        self,
        run_id: str,
        message: str,
        severity: Optional[str] = "INFO",
        tags: Optional[dict[str, str]] = None,
        pretty_print: bool = False,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Create a log for a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :param message: the message to log
        :param severity: the severity of the log
        :param tags: the tags of the log
        :param pretty_print: if True, the log is printed in a pretty format
        :return:
        """
        pass

    def create_webhook(
        self,
        url: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        method: str = "GET",
        headers: Optional[dict] = None,
    ):
        """
        Create a webhook by its name and app id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param url: the url of the webhook
        :param method: the method of the webhook
        :param headers: the headers of the webhook
        """
        pass

    def delete_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_linked_to_templates: Optional[bool] = False,
    ):
        """
        Delete an activity by its name and app id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param with_linked_to_templates: if the activity is linked to a template, delete it anyway
        """
        pass

    def execute_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        run_settings: Union[dict, str] = None,
    ) -> dict:
        """
        Execute an activity by its name or id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_settings: the settings of the run, or the id of the run to clone settings from
        :return: the run of the activity as a dictionary
        """
        pass

    def execute_run(
        self,
        run_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Execute a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return:
        """
        pass

    def get_activities(
        self,
        pretty_print: bool = False,
        print_names: bool = False,
    ) -> list:
        """
        Get the list of activities in the app
        :param pretty_print: if True, the activities are printed in a pretty format
        :param print_names: if True and pretty_print is False, only the names of the activities are printed
        :return: a list of activities as dictionaries
        """
        pass

    def get_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        pretty_print: bool = False,
        how_many_runs: Optional[int] = None,
    ) -> Optional[dict]:
        """
        Get an activity by its name
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param pretty_print: if True, the activity is printed in a pretty format
        :param how_many_runs: how many runs to get
        :return: the activity as a dictionary
        """
        pass

    def get_run_logs(
        self,
        run_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the logs of a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return: the logs of the run as a list of dictionaries
        """
        pass

    def get_run_settings(
        self,
        run_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Get the settings of a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return: the settings of the run as a dictionary
        """
        pass

    def update_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        settings: Optional[dict] = None,
    ):
        """
        Update an activity by its name or id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param new_name: the new name of the activity
        :param settings: the default settings of the activity
        :return: the updated activity as a dictionary
        """
        pass



from typing import TYPE_CHECKING
from shimoku.api.utils import change_data_set_name_with_report
from shimoku.code_gen.business_code_gen.apps_code_gen. \
        data_sets_code_gen.code_gen_from_data_sets import code_gen_read_csv_from_data_set
from shimoku.code_gen.utils_code_gen import code_gen_from_dict, code_gen_from_list

if TYPE_CHECKING:
    from shimoku.api.resources.report import Report
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen


def compact_labels_info(columns: list[dict], data_type_mapping: dict) -> dict:
    """ Compact label info.
    :param columns: columns dict of the table
    :param data_type_mapping: mapping from field to data type
    :return: compacted label info
    """
    label_columns = {}
    for column in columns:
        if 'chips' not in column:
            continue
        chips_dict = column['chips']
        variant = chips_dict['variant']
        mapping_from_field = data_type_mapping[column['field']]
        entry_key = column['field']
        chips_options = chips_dict['options']
        if variant != 'filled':
            entry_key = (entry_key, variant)
        entry_value = {}
        if mapping_from_field.startswith('intField'):
            chips_options = sorted(chips_options, key=lambda x: int(x['value']))
            current_range = (0, chips_options[0]['value'])
            current_color = chips_options[0]['backgroundColor']
            first_index = 0
            i = 1
            for chip in chips_options[1:]:
                if chip['backgroundColor'] != current_color:
                    if i == first_index + 1:
                        entry_value[current_range[1]] = current_color
                    else:
                        entry_value[current_range] = current_color
                    first_index = i
                    current_range = (current_range[1], chip['value'])
                    current_color = chip['backgroundColor']
                else:
                    current_range = (current_range[0], chip['value'])
                i += 1
            if first_index == len(chips_options) - 1:
                entry_value[current_range[1]] = current_color
            else:
                entry_value[(current_range[0], int(current_range[1]) + 1)] = current_color
        else:
            entry_value = {chip['value']: chip['backgroundColor'] for chip in chips_options}
        first_color = chips_options[0]['backgroundColor']
        if all(v == first_color for v in entry_value.values()):
            entry_value = first_color
        label_columns[entry_key] = entry_value
    return label_columns


def get_column_options(columns: list[dict]) -> dict:
    """ Get column options.
    :return: column options
    """
    column_options = {}
    for column in columns:
        not_wanted_keys = ['field', 'headerName', 'order', 'type', 'chips', 'options']
        single_column_options = {k: v for k, v in column.items() if k not in not_wanted_keys}
        if single_column_options:
            column_options[column['field']] = single_column_options
    return column_options


async def code_gen_from_table(
        self: 'AppCodeGen', report: 'Report', properties: dict
) -> list[str]:
    """ Generate code for a table report.
    :param report: report to generate code from
    :param properties: properties of the report
    :return: list of code lines
    """
    report_data_set: Report.ReportDataSet = (await report.get_report_data_sets())[0]
    data_set_id = report_data_set['dataSetId']
    data_set = await self._app.get_data_set(data_set_id)
    data_arg = await code_gen_read_csv_from_data_set(data_set, change_data_set_name_with_report(data_set, report))
    if data_arg is None:
        return ['pass']
    if data_set_id in self._code_gen_tree.shared_data_sets:
        data_arg = f'"{data_set["name"]}",'
    table_params = []
    rev_mapping = {v: (k if data_set['columns'] else v) for k, v in properties['rows']['mapping'].items()}
    mapping = properties['rows']['mapping']
    table_params.append(f'    columns={list(rev_mapping.values())},')
    if properties['pagination']['pageSize'] != 10:
        table_params.append(f'    page_size={properties["pagination"]["pageSize"]},')
    if 'pageSizeOptions' in properties['pagination']:
        page_size_code_lines = code_gen_from_list(properties['pagination']['pageSizeOptions'], 8)
        table_params.extend([f'    page_size_options={page_size_code_lines[0][8:]}',
                             *page_size_code_lines[1:]])
    if not properties['columnsButton']:
        table_params.append(f'    columns_button=False,')
    if not properties['filtersButton']:
        table_params.append(f'    filters=False,')
    if not properties['exportButton']:
        table_params.append(f'    export_to_csv=False,')
    if not properties['search']:
        table_params.append(f'    search=False,')
    if properties.get('sort'):
        sort_field = rev_mapping[properties['sort']['field']]
        sort_direction = properties['sort']['direction']
        table_params.append(f'    initial_sort_column="{sort_field}",')
        if sort_direction != 'asc':
            table_params.append(f'    sort_descending=True,')
    column_options = get_column_options(properties['columns'])
    if column_options:
        column_options_code_lines = code_gen_from_dict(column_options, 8)
        table_params.extend([f'    columns_options={column_options_code_lines[0][8:]}',
                             *column_options_code_lines[1:]])

    categorical_columns = [col_dict['field'] if data_set['columns'] else mapping[col_dict['field']]
                           for col_dict in properties['columns'] if col_dict.get('type') == 'singleSelect']
    if categorical_columns:
        table_params.append(f'    categorical_columns={categorical_columns},')
    label_columns = compact_labels_info(properties['columns'], mapping)
    if label_columns:
        label_columns_code_lines = code_gen_from_dict(label_columns, 4)
        table_params.extend([f'    label_columns={label_columns_code_lines[0][4:]}', *label_columns_code_lines[1:]])
    return [
        'shimoku_client.plt.table(',
        f'    data={data_arg},',
        *self.code_gen_report_params(report.cascade_to_dict()),
        *table_params,
        ')'
    ]



from typing import Optional, Union, TYPE_CHECKING

from copy import deepcopy

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def top_bottom_area_charts(
    self: 'PlotLayer',
    data: Union[str, DataFrame, list[dict]], order: Optional[int] = None,
    x: str = 'x', top_names: Optional[list[str]] = None, bottom_names: Optional[list[str]] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, top_axis_name: Optional[str] = None,
    bottom_axis_name: Optional[str] = None, option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    if top_names is None:
        top_names = ['up']

    if bottom_names is None:
        bottom_names = ['down']

    common_options = get_common_echart_options()
    series_options = get_common_series_options()

    common_options.update({
        'xAxis': [
            {
                'name': x_axis_name if x_axis_name else "",
                'type': 'category',
                'boundaryGap': True,
                'axisLine': {'onZero': False},
                'data': '#set_data#'
            }
        ],
        'yAxis': [
            {
                'name': top_axis_name if top_axis_name else "",
                'type': 'value'
            },
            {
                'name': bottom_axis_name if bottom_axis_name else "",
                'nameLocation': 'start',
                'alignTicks': True,
                'type': 'value',
                'inverse': True
            }
        ],
    })
    series_options.update(
        {
            'type': 'line',
            'areaStyle': {'opacity': 0.5},
            'lineStyle': {
                'width': 1
            },
            'emphasis': {
                'focus': 'series'
            },
            'markArea': {
                'silent': True,
                'itemStyle': {
                    'opacity': 0.3
                },
            },
        }
    )
    series = []
    for name in bottom_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options['name'] = name
        series.append(aux_series_options)

    for name in top_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options['name'] = name
        aux_series_options['yAxisIndex'] = 1
        series.append(aux_series_options)

    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data), order=order, axes=x, values=bottom_names+top_names,
        rows_size=rows_size, cols_size=cols_size, padding=padding, title=title, x_axis_names=x_axis_name,
        y_axis_names=[bottom_axis_name, top_axis_name], echart_options=common_options, series_options=series,
        option_modifications=option_modifications, variant=variant
    )


async def top_bottom_line_charts(
    self: 'PlotLayer',
    data: Union[str, DataFrame, list[dict]], order: int,
    x: str, top_names: Optional[list[str]] = None, bottom_names: Optional[list[str]] = None,
    rows_size: int = 4, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, top_axis_name: Optional[str] = None,
    bottom_axis_name: Optional[str] = None, option_modifications: Optional[dict] = None,
):
    if top_names is None:
        top_names = ['up']

    if bottom_names is None:
        bottom_names = ['down']

    common_options = get_common_echart_options()
    series_options = get_common_series_options()

    common_options.update({
        'grid': [
            {
                'left': 60,
                'right': 50,
                'height': '35%'
            },
            {
                'left': 60,
                'right': 50,
                'top': '55%',
                'height': '35%'
            }
        ],
        'xAxis': [
            {
                'name': x_axis_name if x_axis_name else "",
                'type': 'category',
                'boundaryGap': False,
                'axisLine': {'onZero': True},
                'data': '#set_data#',
                'label': {'show': False}
            },
            {
                'gridIndex': 1,
                'type': 'category',
                'boundaryGap': False,
                'axisLine': {'onZero': True},
                'position': 'top',
                'data': '#set_data#'
            }
        ],
        'yAxis': [
            {
                'name': top_axis_name if top_axis_name else "",
                'type': 'value',
                'nameLocation': 'middle',
                'nameGap': 40,
            },
            {
                'gridIndex': 1,
                'name': bottom_axis_name if bottom_axis_name else "",
                'alignTicks': True,
                'type': 'value',
                'inverse': True,
                'nameLocation': 'middle',
                'nameGap': 40,
            }
        ],
    })
    series_options.update(
        {
            'type': 'line',
            'lineStyle': {
                'width': 1
            },
            'emphasis': {
                'focus': 'series'
            },
            'markArea': {
                'silent': True,
                'itemStyle': {
                    'opacity': 0.3
                },
            },
        }
    )
    series = []
    for name in bottom_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options['name'] = name
        series.append(aux_series_options)

    for name in top_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options['name'] = name
        aux_series_options['yAxisIndex'] = 1
        aux_series_options['xAxisIndex'] = 1
        series.append(aux_series_options)

    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order, axes=[x, x], values=bottom_names + top_names, option_modifications=option_modifications,
        rows_size=rows_size, cols_size=cols_size, padding=padding, title=title, x_axis_names=x_axis_name,
        y_axis_names=[bottom_axis_name, top_axis_name], echart_options=common_options, series_options=series,
    )



import json
from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.utils import (
    CONFIG_PATH, get_profile_config,
    environment_v, access_token_v, universe_id_v, workspace_id_v,
    get_config_profiles, get_current_profile, get_all_configs,
)
from shimoku.exceptions import APIError
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter
from typing import Optional
from rich.console import Console
from rich.table import Table


def add_configuration_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the configuration parser to a parser
    :param parser: Parser to add the configuration parser to
    """

    params = {
        'name': 'config',
        'description': 'Commands to configure the SDK access',
    }

    if parser:
        configuration_parser = parser.add_command(**params)
    else:
        configuration_parser = CLIParser(**params)

    module_functions = [
        switch_profile,
        set,
        get,
        clear,
        list_profiles,
        delete_profile,
    ]

    for func in module_functions:
        configuration_parser.decor_add_command()(func)

    return configuration_parser


def switch_profile(
    profile: str = CLIFuncParam(prompt=True)
):
    """
    Switch the current profile
    :param profile: Profile to switch to
    """
    configs = get_all_configs()
    if profile not in configs['profiles']:
        print(f"The profile {profile} does not exist.")
        return
    with open(CONFIG_PATH, 'w') as f:
        configs['current_profile'] = profile
        json.dump(configs, f)
    print(f"The current profile has been switched to {profile}.")


async def set(
    access_token: Optional[str] = CLIFuncParam(prompt=True),
    universe_id: Optional[str] = CLIFuncParam(prompt=True),
    workspace_id: Optional[str] = CLIFuncParam(prompt=True),
    environment: str = CLIFuncParam(default='production', mandatory=False),
    profile: str = CLIFuncParam(default='default', mandatory=False)
):
    """
    Set the configuration variables for the SDK access
    :param environment: Environment to use
    :param access_token: Access token to use
    :param universe_id: Universe id to use
    :param workspace_id: Workspace id to use
    :param profile: Profile to use
    """
    try:
        await ResourceGetter(InitOptions(
            environment=environment,
            access_token=access_token,
            universe_id=universe_id,
            workspace_id=workspace_id,
        )).get_business()
    except APIError:
        print("The configuration parameters are not valid, please check them and try again.")
        return

    print("The configuration parameters are valid, saving them...")
    configs = get_all_configs()
    current_profile = get_current_profile()
    profile = profile or current_profile
    config = configs.get(profile, {})
    with open(CONFIG_PATH, 'w') as f:
        config[environment_v] = environment if environment else config.get(environment_v, 'production')
        config[access_token_v] = access_token if access_token else config.get(access_token_v, None)
        config[universe_id_v] = universe_id if universe_id else config.get(universe_id_v, 'local')
        config[workspace_id_v] = workspace_id if workspace_id else config.get(workspace_id_v, 'local')

        if 'profiles' not in configs:
            configs['profiles'] = {}
        configs['profiles'][profile] = config
        configs['current_profile'] = profile

        if 'default' not in configs['profiles']:
            configs['profiles']['default'] = config

        json.dump(configs, f)


def get(
    profile: Optional[str]
):
    """
    Get the configuration variables for the SDK access
    :param profile: Profile to use
    """
    if not profile:
        profile = get_current_profile()
    config = get_profile_config(profile)
    table = Table(title=f"{profile[0].upper()+profile[1:].lower()} Configuration Variables")

    for key in config.keys():
        table.add_column(key)
    table_row = [config_value for config_value in config.values()]
    table.add_row(*table_row)
    console = Console()
    console.print(table)


def clear(
    profile: Optional[str]
):
    """
    Clear the configuration variables for the SDK access
    :param profile: Profile to clear
    """
    configs = get_all_configs()
    current_profile = get_current_profile()
    with open(CONFIG_PATH, 'w') as f:
        profile = profile or current_profile
        if profile not in configs['profiles']:
            print(f"The profile {profile} does not exist.")
            return
        configs['profiles'][profile] = {
            environment_v: 'production',
            access_token_v: 'local',
            universe_id_v: 'local',
            workspace_id_v: 'local',
        }
        json.dump(configs, f)
    print("The configuration parameters have been cleared.")


def delete_profile(
    profile: str = CLIFuncParam(prompt=True)
):
    """
    Delete a profile
    :param profile: Profile to delete
    """
    if profile == 'default':
        print("The default profile cannot be deleted.")
        return
    configs = get_all_configs()
    current_profile = get_current_profile()
    configs['profiles'].pop(profile)
    if current_profile == profile:
        configs['current_profile'] = 'default'
        print("The current profile has been switched to default.")
    with open(CONFIG_PATH, 'w') as f:
        json.dump(configs, f)
    print(f"The profile {profile} has been deleted.")


def list_profiles():
    """
    List the profiles
    """
    profiles = get_config_profiles()
    table = Table(title="Profiles")
    table.add_column('Profile')
    table.add_column('Current')
    for profile in profiles:
        table.add_row(profile, 'X' if profile == get_current_profile() else '')
    console = Console()
    console.print(table)



from typing import Optional
import asyncio

from shimoku.api.resources.role import user_delete_role, user_get_role, user_get_roles, user_create_role
from shimoku.utils import create_normalized_name
from shimoku.api.resources.app import App
from shimoku.api.resources.business import Business
from shimoku.api.resources.activity import Activity

import logging
from shimoku.execution_logger import ClassWithLogging
logger = logging.getLogger(__name__)


class MenuPathsLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the Menu Path level.
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(
        self, business: Business,
    ):
        self._business = business
        self._get_for_roles = self._business.get_app if business else None

    async def _get_app_with_warning(
        self, name: Optional[str] = None,
        uuid: Optional[str] = None
    ) -> Optional[App]:
        """
        Get the menu path metadata
        :param name: name of the menu path
        :param uuid: UUID of the menu path
        :return: The menu path object
        """
        app: App = await self._business.get_app(name=name, uuid=uuid, create_if_not_exists=False)
        if not app:
            logger.warning(f"Menu path ({name}) not found")
        return app

    async def get_menu_path(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> dict:
        """
        Get a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :return: menu path metadata
        """
        return (await self._business.get_app(uuid=uuid, name=name)).cascade_to_dict()

    async def update_menu_path(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        new_name: Optional[str] = None, hide_title: Optional[bool] = None,
        hide_path: Optional[bool] = None, show_breadcrumb: Optional[bool] = None,
        show_history_navigation: Optional[bool] = None, order: Optional[int] = None
    ) -> bool:
        """
        Update a menu path
        :param uuid: uuid of the menu path
        :param name: the name of the menu path
        :param new_name: new name of the menu path
        :param hide_title: hide title of the menu path
        :param hide_path: hide path of the menu path
        :param show_breadcrumb: show breadcrumb of the menu path
        :param show_history_navigation: show history navigation of the menu path
        :param order: order of the menu path in the side menu
        :return: menu path metadata
        """
        return await self._business.update_app(
            uuid=uuid, name=name, new_name=new_name,
            normalizedName=create_normalized_name(new_name) if new_name else None,
            hideTitle=hide_title, hidePath=hide_path,
            showBreadcrumb=show_breadcrumb,
            showHistoryNavigation=show_history_navigation,
            order=order
        )

    async def delete_menu_path(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> bool:
        """
        Delete a menu path
        :param uuid: uuid of the menu path
        :param name:
        """
        return await self._business.delete_app(uuid=uuid, name=name)

    async def get_menu_path_activities(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> list[dict]:
        """
        Get the activities of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        app = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return []
        return [activity.cascade_to_dict() for activity in await app.get_activities()]

    async def delete_all_menu_path_activities(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_linked_to_templates: Optional[bool] = False
    ) -> bool:
        """
        Delete all activities of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param with_linked_to_templates: if True, delete all activities, even those linked to templates
        """
        app = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return False
        activities: list[Activity] = await app.get_activities()
        if not with_linked_to_templates:
            previous_len_activities = len(activities)
            activities = [activity for activity in activities if not activity['activityTemplateWithMode']]
            if len(activities) != previous_len_activities:
                logger.warning(
                    "Some activities are linked to templates and will not be deleted. To delete them, "
                    "use the (with_linked_to_templates) parameter set to True."
                )
        await asyncio.gather(*[app.delete_activity(uuid=activity['id']) for activity in activities])
        return True

    async def get_menu_path_components(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> list[dict]:
        """
        Get the reports of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        app = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return []
        return [report.cascade_to_dict() for report in await app.get_reports()]

    async def delete_all_menu_path_components(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None,
        sub_path: Optional[str] = None
    ) -> bool:
        """
        Delete all reports of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param sub_path: sub path of the menu path
        """
        app = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return False
        reports = await app.get_reports()
        if sub_path:
            reports = [report for report in reports if report['path'] == sub_path]
            if not reports:
                logger.warning(f"Sub path ({sub_path}) not found in menu path ({name})")
                return False
        await asyncio.gather(*[app.delete_report(uuid=report['id']) for report in reports])
        return True

    async def get_menu_path_data_sets(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> list[dict]:
        """
        Get the datasets of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        app = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return []
        return [data_set.cascade_to_dict() for data_set in await app.get_data_sets()]

    async def delete_all_menu_path_data_sets(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> bool:
        """
        Delete all datasets of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        app = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return False
        data_sets = await app.get_data_sets()
        await asyncio.gather(*[app.delete_data_set(uuid=data_set['id']) for data_set in data_sets])
        return True

    async def get_menu_path_sub_paths(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None
    ) -> list[str]:
        """
        Get the path names of an menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        app: Optional[App] = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return []
        return await app.get_paths_in_order()

    async def get_menu_path_files(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_shimoku_generated: Optional[bool] = False
    ) -> list[dict]:
        """
        Get the files of an menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param with_shimoku_generated: whether the file is generated internally by the SDK
        """
        app: Optional[App] = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return []
        files = await app.get_files()
        if not with_shimoku_generated:
            files = [file for file in files if 'shimoku_generated' not in file['tags']]
        return [file.cascade_to_dict() for file in files]

    async def delete_all_menu_path_files(
        self, uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_shimoku_generated: Optional[bool] = False
    ) -> bool:
        """
        Delete all files of an menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param with_shimoku_generated: whether the file is generated internally by the SDK
        """
        app: Optional[App] = await self._get_app_with_warning(uuid=uuid, name=name)
        if not app:
            return False
        files = await app.get_files()
        if not with_shimoku_generated:
            previous_len_files = len(files)
            files = [file for file in files if 'shimoku_generated' not in file['tags']]
            if len(files) != previous_len_files:
                logger.warning(
                    "Some files are generated internally by the SDK and will not be deleted. To delete them, "
                    "use the (with_shimoku_generated) parameter set to True."
                )
        await asyncio.gather(*[app.delete_file(uuid=file['id']) for file in files])
        return True

    # Role management
    get_role = user_get_role
    get_roles = user_get_roles
    create_role = user_create_role
    delete_role = user_delete_role



from typing import Optional, List

import tqdm
import subprocess
import os

from shimoku import Business
from shimoku.code_gen.business_code_gen.code_gen_from_business import BusinessCodeGen
from shimoku.code_gen.file_generator import CodeGenFileHandler

from shimoku.utils import create_function_name


async def generate_code(
        business_object: Business,
        business_id: str = 'local',
        access_token: Optional[str] = None,
        universe_id: str = 'local',
        environment: str = 'production',
        output_path: str = 'generated_code',
        menu_paths: Optional[list[str]] = None,
        use_black_formatter: bool = False,
        pbar: Optional[tqdm.tqdm] = None
):
    """ Main code generation function.
    :param business_object: business object to generate code for
    :param business_id: business id to use
    :param access_token: access token to use
    :param universe_id: universe id to use
    :param environment: environment to use
    :param output_path: output path to use
    :param menu_paths: list of menu paths to generate code for
    :param use_black_formatter: whether to use black formatter
    :param pbar: progress bar to use
    """
    await BusinessCodeGen(business_object, output_path).generate_code(business_id, menu_paths, pbar=pbar)
    business_name = create_function_name(business_object['name'])

    main_code_lines: List[str] = [
        'shimoku_client = Client(',
    ]
    if universe_id != 'local':
        main_code_lines.extend([
            f'    access_token="{access_token}",',
            f'    universe_id="{universe_id}",'
        ])
    main_code_lines.extend([
        f'    environment="{environment}",',
        f'    async_execution=True,',
        '    verbosity="INFO"',
        ')',
        '',
        f'workspace_{business_name}(shimoku_client)',
        '',
        'shimoku_client.run()'
    ])
    CodeGenFileHandler(output_path).generate_script_file(
        f'execute_workspace_{business_name}',
        [
            'from shimoku import Client',
            f'from {business_name}.workspace_{business_name} import workspace_{business_name}',
            '',
            '',
            'def main():',
            *[f'    {line}' for line in main_code_lines],
            '',
            '',
            'if __name__ == "__main__":',
            '    main()',
            ''
        ]
    )

    if use_black_formatter:
        # apply black formatting
        subprocess.run(["black", "-l", "80", os.path.join(output_path)])

    if pbar:
        pbar.update(1)



from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_dict
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report


async def code_gen_from_button_modal(
        self: 'AppCodeGen', report: 'Report', report_params: list[str]
) -> list[str]:
    modal_id = report['properties']['events']['onClick'][0]['params']['modalId']
    modal = await self._app.get_report(modal_id)
    path = modal['path']+'_' if modal['path'] else ''
    name = modal["properties"]["hash"][len(path):]
    return [
        'shimoku_client.plt.modal_button(',
        f'    modal="{name}",',
        *report_params,
        ')'
    ]


async def code_gen_from_button_activity(
        self: 'AppCodeGen', report: 'Report', report_params: list[str]
) -> list[str]:
    activity_id = report['properties']['events']['onClick'][0]['params']['activityId']
    activity = await self._app.get_activity(activity_id)
    return [
        'shimoku_client.plt.activity_button(',
        f'    activity_name="{activity["name"]}",',
        *report_params,
        ')'
    ]


async def code_gen_from_button_generic(
        self: 'AppCodeGen', report: 'Report', report_params: list[str]
) -> list[str]:
    events_code = code_gen_from_dict(report['properties']['events'], 4)
    return [
        'shimoku_client.plt.button(',
        *report_params,
        f'    on_click_events={events_code[0][4:]}',
        *events_code[1:],
        ')'
    ]


async def code_gen_from_button(
        self: 'AppCodeGen', report: 'Report'
) -> list[str]:
    """ Generate code for a button report.
    :param report: report to generate code from
    :return: list of code lines
    """
    report_params = [
        f'    label="{report["properties"]["text"]}",',
        f'    order={report["order"]},',
    ]
    if report['sizeColumns'] != 12:
        report_params.append(f'    cols_size={report["sizeColumns"]},')
    if report['sizeRows'] != 1:
        report_params.append(f'    rows_size={report["sizeRows"]},')
    if report['sizePadding'] != '0,0,0,0':
        report_params.append(f'    padding="{report["sizePadding"]}",')
    if report['properties']['align'] != 'stretch':
        report_params.append(f'    align="{report["properties"]["align"]}",')

    if report['properties']['events']['onClick'][0]['action'] == 'openModal':
        return await code_gen_from_button_modal(self, report, report_params)
    elif report['properties']['events']['onClick'][0]['action'] == 'openActivity':
        return await code_gen_from_button_activity(self, report, report_params)
    else:
        return await code_gen_from_button_generic(self, report, report_params)



from typing import Optional, Union, TYPE_CHECKING

from shimoku.api.resources.activity import Activity
from shimoku.exceptions import ActivityError

import json

import logging
from shimoku.execution_logger import log_error, ClassWithLogging
logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from shimoku import App


class ActivitiesLayer(ClassWithLogging):
    """
    This class is used to interact with the activities that are available in a menu path, through the API.
    """
    
    _module_logger = logger
    _use_info_logging = True
    
    def __init__(self, app: Optional['App']):
        self._app = app

    async def _get_activity_with_error(self, uuid: Optional[str], name: Optional[str]):
        """
        Get an activity by its name or id, with error handling
        :param name: the name of the activity
        :param uuid: the id of the activity
        """
        activity: Activity = await self._app.get_activity(uuid=uuid, name=name)
        if not activity:
            log_error(logger, f'Activity ({name}) not found', ActivityError)
        return activity

    async def create_activity(
        self, name: str, settings: Optional[dict] = None,
        template_id: Optional[str] = None, template_name_version: Optional[tuple[str, str]] = None,
        template_mode: str = 'LIGHT', universe_api_key: str = ''
    ) -> dict:
        """
        Create an activity by its name and app id
        :param name: the name of the activity
        :param settings: the settings of the activity
        :param template_id: the template id of the activity
        :param template_name_version: the template name and version of the activity
        :param template_mode: the template mode of the activity
        :param universe_api_key: the universe api key of the activity
        :return: the dictionary representation of the activity
        """
        return (
            await self._app.create_activity(
                name=name, settings=settings, template_id=template_id,
                template_name_version=template_name_version, template_mode=template_mode,
                universe_api_key=universe_api_key
            )
        ).cascade_to_dict()

    async def delete_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None, with_linked_to_templates: Optional[bool] = False
    ):
        """
        Delete an activity by its name and app id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param with_linked_to_templates: if the activity is linked to a template, delete it anyway
        """
        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        if not with_linked_to_templates and activity['activityTemplateWithMode']:
            log_error(
                logger,
                f'Activity ({str(activity)}) is linked to a template please use the '
                f'(with_linked_to_templates) parameter set to True to delete it',
                ActivityError
            )
        await self._app.delete_activity(uuid=uuid, name=name)

    async def update_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        new_name: Optional[str] = None, settings: Optional[dict] = None
    ):
        """
        Update an activity by its name or id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param new_name: the new name of the activity
        :param settings: the default settings of the activity
        :return: the updated activity as a dictionary
        """
        await self._app.update_activity(uuid=uuid, name=name, new_name=new_name, settings=settings)

    async def create_webhook(
        self, url: str,  uuid: Optional[str] = None, name: Optional[str] = None,
        method: str = 'GET', headers: Optional[dict] = None
    ):
        """
        Create a webhook by its name and app id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param url: the url of the webhook
        :param method: the method of the webhook
        :param headers: the headers of the webhook
        """
        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        await activity.create_webhook(url=url, method=method, headers=headers or {})

    async def execute_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        run_settings: Union[dict, str] = None
    ) -> dict:
        """
        Execute an activity by its name or id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_settings: the settings of the run, or the id of the run to clone settings from
        :return: the run of the activity as a dictionary
        """
        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        run = await activity.create_run(settings=run_settings)
        result = await run.trigger_webhook()
        await run
        logger.info(f'Activity {name if name else uuid} executed with result {result}')
        return run.cascade_to_dict()

    async def get_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        pretty_print: bool = False, how_many_runs: Optional[int] = None
    ) -> Optional[dict]:
        """
        Get an activity by its name
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param pretty_print: if True, the activity is printed in a pretty format
        :param how_many_runs: how many runs to get
        :return: the activity as a dictionary
        """
        activity: Activity = await self._app.get_activity(uuid=uuid, name=name)
        if not activity:
            return None

        await activity.get()
        activity_as_dict = activity.cascade_to_dict()

        if how_many_runs:
            activity_as_dict['runs'] = [run.cascade_to_dict() for run in
                                        (await activity.get_runs(how_many_runs=how_many_runs))]
        if pretty_print:
            print(json.dumps(activity_as_dict, indent=2))

        return activity_as_dict

    async def get_activities(
        self, pretty_print: bool = False, print_names: bool = False
    ) -> list[dict]:
        """
        Get the list of activities in the app
        :param pretty_print: if True, the activities are printed in a pretty format
        :param print_names: if True and pretty_print is False, only the names of the activities are printed
        :return: a list of activities as dictionaries
        """

        activities = await self._app.get_activities()
        activities_as_dicts = [activity.cascade_to_dict() for activity in activities]

        if pretty_print:
            print(json.dumps(activities_as_dicts, indent=2))
        elif print_names:
            logger.info(f'Activities available in the app: '
                        f'{", ".join([activity["name"] for activity in activities])}')

        return activities_as_dicts

    async def create_run(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        settings: Union[dict, str] = None
    ) -> dict[str, any]:
        """
        Create a run for an activity by its name
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param settings: the settings of the run, or the id of the run to clone settings from
        :return: the run created as a dictionary
        """

        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        run = await activity.create_run(settings=settings)

        logger.info(f'New run with id {run["id"]} created for activity {name if name else uuid}.')

        return run.cascade_to_dict()

    async def execute_run(
        self, run_id: str, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Execute a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return:
        """
        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        run = await activity.get_run(uuid=run_id)
        result = await run.trigger_webhook()
        await run
        logger.info(f'Run with id {run_id} executed with result {result}')

    async def create_run_log(
        self, run_id: str, message: str, severity: Optional[str] = 'INFO',
        tags: Optional[dict[str, str]] = None, pretty_print: bool = False,
        uuid: Optional[str] = None, name: Optional[str] = None
    ) -> dict[str, str]:
        """
        Create a log for a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :param message: the message to log
        :param severity: the severity of the log
        :param tags: the tags of the log
        :param pretty_print: if True, the log is printed in a pretty format
        :return:
        """
        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        run: Activity.Run = await activity.get_run(uuid=run_id)
        log: dict = (
            await run.create_log(message=message, severity=severity, tags=tags if tags else {})
        ).cascade_to_dict()

        if pretty_print:
            print(json.dumps(log, indent=2))

        return log

    async def get_run_settings(
        self, run_id: str, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> dict[str, any]:
        """
        Get the settings of a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return: the settings of the run as a dictionary
        """

        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        run = await activity.get_run(uuid=run_id)
        return run["settings"]

    async def get_run_logs(
        self, run_id: str, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict[str, any]]:
        """
        Get the logs of a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return: the logs of the run as a list of dictionaries
        """

        activity: Activity = await self._get_activity_with_error(uuid=uuid, name=name)
        run = await activity.get_run(uuid=run_id)
        return [log.cascade_to_dict() for log in await run.get_logs()]



""""""
import json
import time
from os import getenv
from typing import Dict
from unittest import TestCase

import pandas as pd

from shimoku import Client
from shimoku.exceptions import (DataError)

access_token: str = getenv('API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
verbosity: str = getenv('VERBOSITY')


s = Client(
    access_token=access_token,
    universe_id=universe_id,
    verbosity=verbosity
)
s.set_workspace(uuid=business_id)
s.menu_paths.delete_menu_path(name='data_set_test')
s.set_menu_path('data_set_test')
# s.plt.clear_menu_path()

# Fixtures
data: Dict = {
    'a': [x for x in range(10)],
    'b': [x ** 2 for x in range(10)],
}
data_oriented = sorted(pd.DataFrame(data).to_dict(orient='records'), key=lambda x: x['a'])
df = pd.DataFrame(data)
data_json: str = json.dumps(data_oriented)


def sleep_cloud():
    if not s.playground:
        time.sleep(5)


def test_data_sets():
    global data_oriented
    s.data.append_to_data_set(name='test-data-set', data=data_oriented)
    sleep_cloud()
    data_ = s.data.get_data_from_data_set(name='test-data-set')
    data_ = sorted(data_, key=lambda x: x['intField1'])
    for i in range(len(data_oriented)):
        assert data_[i]['intField1'] == data_oriented[i]['a']
        assert data_[i]['intField2'] == data_oriented[i]['b']

    s.data.append_to_data_set(name='test-data-set', data=data_oriented)
    sleep_cloud()
    data_ = s.data.get_data_from_data_set(name='test-data-set')
    data_ = sorted(data_, key=lambda x: x['intField1'])
    for i in range(len(data_oriented)):
        assert data_[i*2]['intField1'] == data_oriented[i]['a']
        assert data_[i*2]['intField2'] == data_oriented[i]['b']

    s.data.replace_data_from_data_set(name='test-data-set', data=data_oriented)
    sleep_cloud()
    data_ = s.data.get_data_from_data_set(name='test-data-set')
    data_ = sorted(data_, key=lambda x: x['intField1'])
    assert len(data_) == len(data_oriented)
    for i in range(len(data_oriented)):
        assert data_[i]['intField1'] == data_oriented[i]['a']
        assert data_[i]['intField2'] == data_oriented[i]['b']

    s.data.delete_data_set(name='test-data-set')
    assert not s.data.get_data_from_data_set(name='test-data-set')


class TestBadDfs(TestCase):

    bad_dfs = [
        [{'a': 1, 'b': 2}, {'a': 1, 'b': 2, 'c': 3}],
        [{'a': 1, 'b': 2}, {'a': 1, 'b': 'c'}],
        [{'a': 1, 'b': 2}, {'a': None, 'b': 2}],
    ]

    def test_bad_dfs(self):
        for i, bad_df in enumerate(self.bad_dfs):
            with self.assertRaises(DataError):
                s.data.append_to_data_set(name=f'test-data-set_{i}', data=bad_df)

    def test_bad_append_to_existing_df(self):
        s.data.replace_data_from_data_set(name='test-data-set', data=data_oriented)
        sleep_cloud()
        print(s.data.get_data_from_data_set(name='test-data-set'))
        for i, bad_df in enumerate(self.bad_dfs):
            bad_df_aux = [bad_df[1]]
            with self.assertRaises(DataError):
                s.data.append_to_data_set(name=f'test-data-set', data=bad_df_aux)


test_data_sets()

test_bad_dfs = TestBadDfs()
test_bad_dfs.test_bad_dfs()
test_bad_dfs.test_bad_append_to_existing_df()



from os import getenv
from shimoku import Client
# from mock_classes import MockClient


def initiate_shimoku():
    api_key: str = getenv('API_TOKEN')
    universe_id: str = getenv('UNIVERSE_ID')
    environment: str = getenv('ENVIRONMENT')
    verbose: str = getenv('VERBOSITY')
    mock: bool = getenv('MOCK') == 'TRUE'
    async_execution: bool = getenv('ASYNC_EXECUTION') == 'TRUE'
    playground: bool = getenv('PLAYGROUND') == 'TRUE'
    local_port: int = int(getenv('LOCAL_PORT'))

    if playground:
        s = Client(
            environment=environment,
            verbosity=verbose,
            async_execution=async_execution,
            local_port=local_port,
            retry_attempts=1
        )
    # elif mock:
        # s = MockClient(
        #     verbosity=verbose,
        #     async_execution=async_execution,
        # )
    else:
        s = Client(
            access_token=api_key,
            universe_id=universe_id,
            environment=environment,
            verbosity=verbose,
            async_execution=async_execution,
            retry_attempts=1
        )

    return s



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class AIFunctionMethodsHeader:
    """None"""

    def create_model(
        self,
        model_name: str,
        model: bytes,
        metadata: Optional[dict] = None,
    ):
        """ Create a model
        :param model_name: Name of the model
        :param model: Model to be uploaded
        :param metadata: Metadata of the model
        """
        pass

    def create_output_files(
        self,
        files: dict,
        model_name: Optional[str] = None,
    ):
        """ Create output files of a ai_function
        :param files: Files to be uploaded
        :param model_name: Name of the model used
        """
        pass

    def get_model(
        self,
        model_name: str,
    ) -> tuple:
        """ Get a model
        :param model_name: Name of the model
        :return: The model if it exists
        """
        pass


class AILayerHeader:
    """None"""

    def check_for_private_access(
        self,
        ai_function_id: str,
        run_id: str,
    ):
        """ Check if the credentials allow for private ai_function methods """
        pass

    def create_input_files(
        self,
        input_files: dict,
        force_overwrite: bool = False,
    ):
        """ Create input files for a ai_function
        :param input_files: dictionary of input files to create
        :param force_overwrite: Force the overwrite of the files
        """
        pass

    def delete_input_file(
        self,
        file_name: str,
    ):
        """ Delete an input file
        :param file_name: Name of the file to delete
        """
        pass

    def delete_model(
        self,
        model_name: str,
    ):
        """ Delete a model
        :param model_name: Name of the model
        """
        pass

    def generic_execute(
        self,
        ai_function: str,
        version: Optional[str] = None,
        universe_api_key: Optional[str] = None,
        **kwargs,
    ) -> str:
        """ Execute a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param universe_api_key: API key of the universe
        :param params: Parameters to be passed to the ai_function
        """
        pass

    def get_available_input_files(
        self,
    ) -> list:
        """ Get available input files for a ai_function """
        pass

    def get_last_executions_with_logs(
        self,
        ai_function: str,
        version: Optional[str] = None,
        how_many: int = 1,
    ):
        """ Get the logs of the executions of a ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param how_many: Number of executions to get
        :return: The logs of the ai_function
        """
        pass

    def get_output_file_objects(
        self,
        run_id: str,
        file_name: Optional[str] = None,
    ) -> Union[tuple[bytes, dict], dict[str, tuple[bytes, dict]]]:
        """ Get an output file object for a given ai_function and run id
        :param run_id: ID of the executed run
        :param file_name: Name of the file to get
        """
        pass

    def get_output_files_by_ai_function(
        self,
        ai_function: str,
        version: Optional[str] = None,
        file_name: Optional[str] = None,
        get_objects: bool = False,
    ):
        """ Get output files for a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :return: The output files if they exist
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        """
        pass

    def get_output_files_by_model(
        self,
        model_name: str,
        file_name: Optional[str] = None,
        get_objects: bool = False,
    ):
        """ Get output files for the executions of ai_functions with a given model
        :param model_name: Name of the model to use
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        :return: dictionary of output files by the execution identifier
        """
        pass

    def get_private_ai_function_methods(
        self,
    ) -> AIFunctionMethodsHeader:
        """ Get the private ai_function methods if the credentials allow for it """
        pass



from typing import Optional, TYPE_CHECKING, Type

import asyncio
from copy import deepcopy

from shimoku.api.base_resource import Resource
from shimoku.api.resources.data_set import DataSet, Mapping
if TYPE_CHECKING:
    from shimoku.api.resources.app import App

import logging
logger = logging.getLogger(__name__)


class Report(Resource):
    """ Report resource class """

    _module_logger = logger
    resource_type = 'report'
    plural = 'reports'
    alias_field = 'properties/hash'

    report_type = None
    possible_values = {}
    default_properties = {
        'hash': None,
    }

    class ReportDataSet(Resource):

        _module_logger = logger
        resource_type = 'reportDataSet'
        plural = 'reportDataSets'

        def __init__(self, parent: 'Report', uuid: Optional[str] = None, db_resource: Optional[dict] = None):

            params = dict(
                reportId=parent['id'],
                dataSetId='',
                properties={},
            )

            super().__init__(parent=parent, uuid=uuid, db_resource=db_resource, params=params,
                             check_params_before_creation=['reportId', 'dataSetId'],
                             params_to_serialize=['properties'])

        async def delete(self):
            """ Delete the report dataset """
            return await self._base_resource.delete()

        async def update(self):
            """ Update the report dataset """
            return await self._base_resource.update()

    def __init__(self, parent: 'App', uuid: Optional[str] = None, db_resource: Optional[dict] = None):

        params = dict(
            title=None,
            path=None,
            pathOrder=None,
            reportType=self.report_type,
            order=0,
            sizeColumns=12,
            sizeRows=3,
            sizePadding='0,0,0,0',
            bentobox={},
            properties=deepcopy(self.default_properties),
            dataFields={},
            chartData=[],
            # subscribed=False,
            smartFilters=[],
        )

        super().__init__(parent=parent, uuid=uuid, db_resource=db_resource, params=params,
                         children=[Report.ReportDataSet],
                         check_params_before_creation=['order'],
                         params_to_serialize=['properties', 'dataFields', 'chartData', 'bentobox', 'smartFilters'])

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other: 'Report'):
        return super().__eq__(other)

    def __new__(cls, parent: 'App', uuid: Optional[str] = None, db_resource: Optional[dict] = None):
        if uuid:
            return super().__new__(cls)
        if cls is Report:
            if db_resource is None:
                raise ValueError('You must provide a db_resource to create a Report instance')
            if db_resource['reportType'] == 'TABS':
                from .reports.tabs_group import TabsGroup
                return TabsGroup(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'MODAL':
                from .reports.modal import Modal
                return Modal(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'INDICATOR':
                from .reports.charts.indicator import Indicator
                return Indicator(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'ECHARTS2':
                from .reports.charts.echart import EChart
                return EChart(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'IFRAME':
                from .reports.charts.iframe import IFrame
                return IFrame(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'HTML':
                from .reports.charts.html import HTML
                return HTML(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'TABLE':
                from .reports.charts.table import Table
                return Table(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'ANNOTATED_ECHART':
                from .reports.charts.annotated_chart import AnnotatedEChart
                return AnnotatedEChart(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'BUTTON':
                from .reports.charts.button import Button
                return Button(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'FORM':
                from .reports.charts.input_form import InputForm
                return InputForm(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] == 'FILTERDATASET':
                from .reports.filter_data_set import FilterDataSet
                return FilterDataSet(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource['reportType'] in ['INDICATORS', 'MULTIFILTER', 'ECHARTS', None]:
                from .reports.unsupported import Unsupported
                return Unsupported(parent=parent, uuid=uuid, db_resource=db_resource)
            else:
                raise ValueError(f'Unknown report type {db_resource["reportType"]}')
        else:
            return super().__new__(cls)

    async def delete(self):
        """ Delete the report """
        return await self._base_resource.delete()

    async def update(self):
        """ Update the report """
        return await self._base_resource.update()

    def set_properties(self, **properties):
        """ Set the properties of the report without saving it to the server ç
        :param properties: the properties to set """
        for property_name, value in properties.items():
            if property_name not in self.default_properties.keys():
                raise ValueError(f'Property {property_name} is not a possible property for {self.report_type}')
            if property_name in self.possible_values and value not in self.possible_values[property_name]:
                raise ValueError(f'Value {value} is not a possible value for property {property_name}')

        self['properties'].update(properties)
        self._base_resource.changed_params.add('properties')

    async def change_report_type(self, report_class: Type['Report']):
        """ change the report type of the report """
        logger.info(f'Changing component type from {self.report_type} to {report_class.report_type}')
        if self.report_type == report_class.report_type:
            return
        await self.delete_report_data_sets()
        self['reportType'] = report_class.report_type
        r_hash = self['properties'].get('hash')
        self['properties'] = deepcopy(report_class.default_properties)
        self['properties']['hash'] = r_hash
        self['dataFields'] = {}
        self['chartData'] = []
        self.__class__ = report_class

    # ReportDataSet methods
    async def create_report_data_set(self, mapping_data_set_sort: tuple[Optional[Mapping], DataSet, Optional[dict]],
                                     properties: Optional[dict] = None) -> 'ReportDataSet':
        """ Update the report dataset
        :param mapping_data_set_sort: a tuple of mapping, data_set, sort
        :param properties: the properties to set
        :return: the updated report dataset
        """
        mapping, data_set, sort = mapping_data_set_sort
        data_set_id = data_set['id']
        properties = properties if properties is not None else {}
        properties['mapping'] = mapping
        properties['sort'] = sort
        params = dict(
            dataSetId=data_set_id,
            properties=properties,
        )
        report_data_set = await self._base_resource.create_child(self.ReportDataSet, **params)
        return report_data_set

    async def get_report_dataset(self, uuid: Optional[str] = None) -> 'ReportDataSet':
        """ Get a report dataset
        :param uuid: the uuid of the report dataset to get
        :return: the report dataset
        """
        return await self._base_resource.get_child(self.ReportDataSet, uuid=uuid)

    async def get_report_data_sets(self) -> list['ReportDataSet']:
        """ Get all the report datasets of the report """
        return await self._base_resource.get_children(self.ReportDataSet)

    async def delete_report_dataset(self, uuid: Optional[str] = None) -> bool:
        """ Delete a report dataset
        :param uuid: the uuid of the report dataset to delete
        :return: True if the report dataset was deleted, False otherwise
        """
        return await self._base_resource.delete_child(self.ReportDataSet, uuid=uuid)

    async def delete_report_data_sets(self, log: bool = False):
        """ Delete the datasets and links that are not in use anymore.
        :param log: if True, log the actions"""
        report_data_sets = await self.get_report_data_sets()
        await asyncio.gather(*[self._base_resource.delete_child(self.ReportDataSet, uuid=rds['id'])
                               for rds in report_data_sets])
        if log:
            logger.info(f'Deleted {len(report_data_sets)} component data set links from component at {str(self)}')



""""""
from os import getenv
from typing import Dict, List

from shimoku import Client


api_key: str = getenv('API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')

config = {
    'access_token': api_key,
}

s = Client(
    config=config,
    universe_id=universe_id,
    verbosity='DEBUG',
)


def test_get_universe_businesses():
    workspaces: List[Dict] = s.universes.get_universe_workspaces(uuid=universe_id)
    assert workspaces


test_get_universe_businesses()



from typing import Optional, TYPE_CHECKING

from shimoku.api.base_resource import Resource
if TYPE_CHECKING:
    from shimoku.api.resources.business import Business

import logging
logger = logging.getLogger(__name__)


class Event(Resource):
    """ Event resource class """

    _module_logger = logger
    resource_type = 'event'
    plural = 'events'

    def __init__(self, parent: 'Business', uuid: Optional[str] = None):

        params = dict(
            type='NO_EVENT',
            resourceId=None,
            content={},
        )

        super().__init__(
            parent=parent, uuid=uuid, params=params,
            params_to_serialize=['content']
        )



""""""
from os import getenv
from typing import Optional
import unittest
import time
import json
import datetime as dt
from shimoku.exceptions import ShimokuFileError
from shimoku import Client

access_token: str = getenv('UNIVERSE_API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
verbosity: str = getenv('VERBOSITY')
environment: str = getenv('ENVIRONMENT')

last_execution_time: float = 0


def execute_ai_function(wf: str, model_name: Optional[str] = None) -> str:
    global last_execution_time
    tics = 0
    while last_execution_time is not None and time.time() - last_execution_time < 15:
        print('\rWaiting for min run interval'+('.'*(tics % 4))+'      ', end='')
        tics += 1
        time.sleep(0.5)
    print()
    s.ai.create_input_files(
        input_files={
            f'Great Content': (
                b'//////////////////////////// CONTENT OF THE INPUT FILE ////////////////////////////\n\n'
                b'Great Content\n\n///////////////////////////////////////////////////////////////////////////////////',
                {'Great metadata entry': 'important file info'}
            )
        },
        force_overwrite=True
    )
    last_execution_time = time.time()
    params = dict(
        ai_function=wf,
        input_data='Great Content',
        text_mandatory='Important text',
        text_optional='Less important text',
        add_timestamp=True
    )
    if model_name:
        params['model_name'] = model_name
    return s.ai.generic_execute(**params)


s = Client(
    access_token=access_token,
    universe_id=universe_id,
    verbosity=verbosity,
    environment=environment,
    retry_attempts=1
)
s.set_workspace(uuid=business_id)

AI_FUNCTION_TO_TEST = 'test_sdk_workflow_create_outputs'
AI_FUNCTION_ID_TO_TEST = s.activity_templates.get_activity_template(name_version=(AI_FUNCTION_TO_TEST, ''))['id']
MENU_PATH = 'test-ai'
MODEL_NAME = 'test-model'

if MENU_PATH in {menu['name'] for menu in s.workspaces.get_workspace_menu_paths(business_id)}:
    s.menu_paths.delete_all_menu_path_files(name=MENU_PATH, with_shimoku_generated=True)
    s.menu_paths.delete_all_menu_path_activities(name=MENU_PATH, with_linked_to_templates=True)
    s.menu_paths.delete_menu_path(name=MENU_PATH)

s.set_menu_path(name=MENU_PATH)


def test_private_methods():
    run_id_for_model = execute_ai_function(AI_FUNCTION_TO_TEST)
    s.ai.check_for_private_access(
        ai_function_id=AI_FUNCTION_ID_TO_TEST, run_id=run_id_for_model
    )
    ai_function_methods = s.ai.get_private_ai_function_methods()
    ai_function_methods.create_model(
        model_name=MODEL_NAME, model=b'',
        metadata={'other_field': 'other_value'}
    )
    model, metadata = ai_function_methods.get_model(model_name=MODEL_NAME)
    assert model == b''
    assert metadata == {
        'creator_ai_function': AI_FUNCTION_TO_TEST,
        'creator_ai_function_id': AI_FUNCTION_ID_TO_TEST,
        'creator_ai_function_version': '',
        'model_name': MODEL_NAME,
        'other_field': 'other_value',
        'run_id': run_id_for_model
    }
    s.ai.delete_model(model_name='test-model')

    class ModelDoesntExist(unittest.TestCase):
        def test_model_doesnt_exist(self):
            with self.assertRaises(ShimokuFileError):
                ai_function_methods.get_model(model_name=MODEL_NAME)

    t = ModelDoesntExist()
    t.test_model_doesnt_exist()

    # For the following tests the model must exist
    ai_function_methods.create_model(
        model_name=MODEL_NAME, model=b'',
        metadata={'other_field': 'other_value'}
    )


def test_create_input_files():
    input_files = s.ai.get_available_input_files()
    if 'file1' in [input_file['file_name'] for input_file in input_files]:
        s.ai.delete_input_file(file_name='file1')
    if 'file2' in [input_file['file_name'] for input_file in input_files]:
        s.ai.delete_input_file(file_name='file2')
    len_files = len(s.menu_paths.get_menu_path_files(name=MENU_PATH))
    len_input_files = len(s.ai.get_available_input_files())
    s.ai.create_input_files({'file1': b'', 'file2': (b'', {'metadata_key': 'metadata_value'})})
    assert len(s.menu_paths.get_menu_path_files(name=MENU_PATH)) == len_files
    input_files = s.ai.get_available_input_files()
    assert len(input_files) == len_input_files + 2
    assert 'file1' in [input_file['file_name'] for input_file in input_files]
    assert 'file2' in [input_file['file_name'] for input_file in input_files]
    assert (('metadata_key', 'metadata_value') in
            [input_file_metadata.items()
                for input_file_metadata in input_files if input_file_metadata['file_name'] == 'file2'][0])
    s.ai.delete_input_file(file_name='file1')
    s.ai.delete_input_file(file_name='file2')
    assert len(s.menu_paths.get_menu_path_files(name=MENU_PATH, with_shimoku_generated=True)) == len_files + 2
    assert len(s.ai.get_available_input_files()) == len_input_files


def test_execute_ai_function():
    run_id = execute_ai_function(AI_FUNCTION_TO_TEST, MODEL_NAME)
    first_time = time.time()
    activity_ended = False
    last_log = ''
    s.disable_caching()
    while time.time() - first_time < 60:
        run = s.ai.get_last_executions_with_logs(AI_FUNCTION_TO_TEST)[0]
        if last_log != run['logs'][-1]['message']:
            last_log = run['logs'][-1]['message']
            print(last_log)
            if last_log == 'The Activity has ended':
                activity_ended = True
                break
    if not activity_ended:
        raise Exception('AI function did not finish')
    output_dict = s.ai.get_output_file_objects(run_id=run_id)
    s.enable_caching()
    file_name, (file_obj, metadata) = list(output_dict.items())[0]

    decoded_file = file_obj.decode('utf-8')
    assert 'Important text' in decoded_file
    assert 'Less important text' in decoded_file
    assert 'Great Content' in decoded_file
    assert dt.datetime.now().strftime('%Y-%m-%d') in file_name
    assert metadata['model_name'] == MODEL_NAME

    def assert_output_dict(_output_dict: dict):
        if 'model_name' in _output_dict:
            assert _output_dict['model_name'] == MODEL_NAME
        if 'ai_function_name' in _output_dict:
            assert _output_dict['ai_function_name'] == AI_FUNCTION_TO_TEST
            assert _output_dict['ai_function_version'] == ""

        assert _output_dict['input']['args'] == {
            'add_timestamp': True, 'text_mandatory': 'Important text',
            'text_optional': 'Less important text', 'model_name': MODEL_NAME
        }
        assert _output_dict['input']['files'] == {
            'input_data': {
                'file_name': 'Great Content', 'Great metadata entry': 'important file info'
            }
        }
        assert _output_dict['output_files'][file_name]['model_name'] == MODEL_NAME

    output_dict_by_ai_function = [
        o_dict for o_dict in s.ai.get_output_files_by_ai_function(AI_FUNCTION_TO_TEST) if o_dict['run_id'] == run_id
    ][0]
    assert_output_dict(output_dict_by_ai_function)
    output_dict_by_model = [
        o_dict for o_dict in s.ai.get_output_files_by_model(MODEL_NAME) if o_dict['run_id'] == run_id
    ][0]
    assert_output_dict(output_dict_by_model)


test_private_methods()
test_create_input_files()
test_execute_ai_function()

s.ai.show_available_ai_functions()
s.ai.show_ai_function_parameters(AI_FUNCTION_TO_TEST)
s.ai.show_available_models()
s.ai.show_available_input_files()
print(json.dumps(s.ai.get_output_files_by_ai_function(AI_FUNCTION_TO_TEST), indent=4))
s.ai.show_last_execution_logs_by_ai_function(AI_FUNCTION_TO_TEST)
s.ai.show_last_execution_logs_by_model(MODEL_NAME)



from typing import Optional, Union, TYPE_CHECKING

from copy import deepcopy

import pandas as pd

from shimoku.plt.utils import ShimokuPalette
from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


async def line_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    """ Create a line chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options['type'] = 'line'
    common_options['yAxis'][0]['scale'] = True
    common_options['xAxis'][0]['boundaryGap'] = False
    if variant == 'minimal':
        series_options['itemStyle'] = {'opacity': 0}
        series_options['lineStyle'] = {'width': 3}

    await self._create_trend_chart(
        option_modifications=option_modifications,
        echart_options=common_options, series_options=series_options, variant=variant,
        axes=x, values=y, data_mapping_to_tuples=await self._choose_data(order, data),
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values
    )


async def area_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    """ Create an area chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options.update({
        'type': 'line',
        'areaStyle': {'opacity': 0.5},
    })
    if variant == 'minimal':
        series_options['itemStyle'] = {'opacity': 0}
        series_options['areaStyle']['opacity'] = 0.2
        series_options['lineStyle'] = {'width': 3}

    common_options['xAxis'][0]['boundaryGap'] = False
    await self._create_trend_chart(
        option_modifications=option_modifications,
        echart_options=common_options, series_options=series_options,
        axes=x, values=y, data_mapping_to_tuples=await self._choose_data(order, data),
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values,
        variant=variant
    )


async def stacked_area_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None, option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    """ Create a stacked area chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options.update({
        'type': 'line',
        'areaStyle': {},
        'stack': 'stack',
        'itemStyle': {'borderRadius': [0, 0, 0, 0]},
    })
    common_options['xAxis'][0]['boundaryGap'] = False
    data_mapping_to_tuples = await self._choose_data(order, data)
    len_y = len(y) if y else len(data_mapping_to_tuples.keys())-1
    series_options = [series_options] * (len_y-1)
    series_options.append(deepcopy(series_options[-1]))
    series_options[-1]['itemStyle']['borderRadius'] = [9, 9, 0, 0]
    await self._create_trend_chart(
        option_modifications=option_modifications,
        echart_options=common_options, series_options=series_options,
        axes=x, values=y, data_mapping_to_tuples=data_mapping_to_tuples,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values,
        variant=variant
    )


async def predictive_line_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, y: Union[list[str], str, None] = None,
    min_value_mark: Optional[Union[str, int]] = None,
    max_value_mark: Optional[Union[str, int]] = None,
    color_mark: str = 'rgba(255, 173, 177, 0.4)',
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    title: Optional[str] = None, rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None, variant: Optional[str] = None
):
    """ Create a predictive line chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options['type'] = 'line'
    common_options['yAxis'][0]['scale'] = True
    common_options['xAxis'][0]['boundaryGap'] = False
    series_options['markArea'] = {
        'itemStyle': {'color': color_mark},
        'data': [
            [
                {'name': 'Prediction', 'xAxis': min_value_mark},
                {'xAxis': max_value_mark},
            ]
        ]
    }
    await self._create_trend_chart(
        echart_options=common_options, series_options=series_options, option_modifications=option_modifications,
        axes=x, values=y, data_mapping_to_tuples=await self._choose_data(order, data),
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values, variant=variant
    )


async def marked_line_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, marks: list,  y: Union[list[str], str, None] = None,
    color_mark: str = 'rgba(255, 173, 177, 0.4)',
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    title: Optional[str] = None, rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None, variant: Optional[str] = None
):
    """ Create a predictive line chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options['type'] = 'line'
    common_options['yAxis'][0]['scale'] = True
    common_options['xAxis'][0]['boundaryGap'] = False
    series_options['markArea'] = {
        'itemStyle': {'color': color_mark},
        'data': [[
            {'name': name, 'xAxis': min_value_mark},
            {'xAxis': max_value_mark},
        ] for name, min_value_mark, max_value_mark in marks]
    }
    await self._create_trend_chart(
        echart_options=common_options, series_options=series_options, option_modifications=option_modifications,
        axes=x, values=y, data_mapping_to_tuples=await self._choose_data(order, data),
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values, variant=variant
    )


def calculate_segments(df: pd.DataFrame, threshold: float, r: int, g: int, b: int):
    color_range = df['y'].max() - threshold
    df_as_dict = df.to_dict('records')
    segment = [None, None, None]
    segments = []
    for i in range(len(df_as_dict)):
        if df_as_dict[i]['y'] >= threshold and segment[0] is None:
            segment[0] = i-1 if i > 0 else 0
        elif (df_as_dict[i]['y'] < threshold or i == len(df_as_dict) - 1) and segment[0] is not None:
            if i < len(df_as_dict) - 1 and df_as_dict[i+1]['y'] >= threshold:
                continue
            segment[1] = i
            max_from_segment = max([val['y'] for val in df_as_dict[segment[0]:segment[1]]])
            color_alpha = 0.3 + (0.7 * (max_from_segment - threshold) / color_range if color_range > 0 else 0)
            segment[2] = f'rgba({r}, {g}, {b}, {color_alpha})'
            segments.append(tuple(segment))
            segment = [None, None, None]

    return segments


async def segmented_area_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, y: str, segments: Optional[list] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    default_color: tuple[int, int, int] = (255, 0, 0), title: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None, variant: Optional[str] = None,
    top_area: bool = False, threshold: Optional[float] = None, labels: Optional[list[str]] = None
):
    """ Create a segmented area chart """
    if not isinstance(y, str):
        raise TypeError('y must be string')
    if threshold is not None and segments is not None:
        raise ValueError('You cannot specify both threshold and segments')
    if threshold is None and segments is None:
        raise ValueError('You must specify either threshold or segments')
    data_mappings_to_tuples = await self._choose_data(order, data)
    if isinstance(data, str):
        data = self._shared_data[data]
    df = pd.DataFrame(data)
    if threshold is not None:
        segments = calculate_segments(df, threshold, *default_color)

    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    common_options['xAxis'][0]['boundaryGap'] = False
    common_options['legend'] = {'show': False}
    del common_options['toolbox']['feature']['magicType']
    del common_options['toolbox']['bottom']
    common_options['toolbox']['top'] = "0px"
    common_options['dataZoom'] = [{'type': 'inside'}, {'type': 'slider'}] if variant is None else []
    pieces = []
    for segment in segments:
        if len(segment) == 2:
            r, g, b = default_color
            pieces.append({'gte': segment[0], 'lte': segment[1], 'color': f'rgba({r}, {g}, {b}, 0.75)'})
        elif len(segment) == 3:
            color = segment[2]
            if not isinstance(color, str):
                if not isinstance(color, tuple) or len(color) != 3:
                    log_error(logger, f'Color must be a string or a tuple of 3 elements, got {color}', ValueError)
                color = f'rgba({color[0]}, {color[1]}, {color[2]}, 0.75)'
            pieces.append({'gte': segment[0], 'lte': segment[1], 'color': color})
        else:
            raise ValueError('Segment must be a tuple of 2 or 3 elements')
    common_options['visualMap'] = {
        'type': 'piecewise',
        'show': False,
        'dimension': 0,
        'seriesIndex': 0,
        'pieces': pieces
    }
    series_options.update({
        'type': 'line',
        'emphasis': {'disabled': True},
        'areaStyle': {'origin': 'end' if top_area else 'start'},
        'itemStyle': {'opacity': 0},
    })

    series_options = [series_options, deepcopy(series_options)]
    del series_options[1]['areaStyle']

    if labels is None:
        labels = []
    if len(labels) < len(segments):
        labels += [''] * (len(segments) - len(labels))
    if len(labels) > len(segments):
        labels = labels[:len(segments)]

    series_options[1]['markArea'] = {
        'itemStyle': {'color': 'rgba(0, 0, 0, 0)'},
        'emphasis': {'disabled': True},
        'label': {
            'show': variant != 'minimal',
            'emphasis': {'disabled': True},
            'borderColor': f'rgba({default_color[0]}, {default_color[1]}, {default_color[2]}, 1)',
            'borderWidth': 1,
            'borderRadius': 5,
            'padding': [5, 5, 5, 5],
            'backgroundColor': f'rgba(0,0,0,0)'
        },
        'data': [[
            {'name': label, 'xAxis': marks_tuple[0]},
            {'xAxis': marks_tuple[1]},
        ] for label, marks_tuple in zip(labels, segments) if label]
    }

    await self._create_trend_chart(
        echart_options=common_options, series_options=series_options, option_modifications=option_modifications,
        axes=x, values=[y, y], data_mapping_to_tuples=data_mappings_to_tuples,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values, variant=variant
    )


async def line_with_confidence_area_chart(
    self: 'PlotLayer', data: Union[str, pd.DataFrame, list[dict]],
    order: int, x: str, lower: str, y: str, upper: str,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    percentages: bool = False, option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    common_options = get_common_echart_options()
    del common_options['toolbox']['feature']['magicType']
    common_options.update({
        'tooltip': {
            'trigger': 'axis',
            'axisPointer': {
                'type': 'cross',
                'animation': False,
                'label': {
                    'backgroundColor': '#ccc',
                    'borderColor': '#aaa',
                    'borderWidth': 1,
                    'shadowBlur': 0,
                    'shadowOffsetX': 0,
                    'shadowOffsetY': 0,
                    'color': '#222'
                }
            },
        },
        'xAxis': [
            {
                'type': 'category',
                'fontFamily': 'Rubik',
                'nameLocation': 'middle',
                'boundaryGap': True,
                'nameGap': 35,
                'data': '#set_data#'
            }
        ],
        'yAxis': [
            {
                'type': 'value',
                'fontFamily': 'Rubik',
                'splitNumber': 3,
                'boundaryGap': True,
                'nameLocation': 'middle',
                'nameGap': 60,
                'axisLabel': {
                    'formatter': '{value}%' if percentages else '{value}'
                },
                'axisPointer': {
                    'label': {
                        'formatter': '{value}%' if percentages else '{value}'
                    }
                },
            },
        ],
    })
    series_options = [
        {
            'name': lower,
            'data': '#set_data#',
            'type': 'line',
            'lineStyle': {
                'color': '#000',
                'opacity': 0.1
            },
            'areaStyle': {
                'color': '#000',
                'opacity': 0.1,
                'origin': 'end'
            },
            'symbol': 'none'
        },
        {
            'name': upper,
            'data': '#set_data#',
            'silent': True,
            'type': 'line',
            'lineStyle': {
                 'color': '#000',
                 'opacity': 0.1
            },
            'areaStyle': {
                'color': '#000',
                'opacity': 0.1,
                'origin': 'start'
            },
            'symbol': 'none'
        },
        {
            'name': y,
            'data': '#set_data#',
            'type': 'line',
            'itemStyle': {
                'color': ShimokuPalette.CHART_C1.value
            },
            'emphasis': {
                'lineStyle': {
                    'color': ShimokuPalette.CHART_C1.value
                }
            },
            'showSymbol': False
        }
    ]

    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data), order=order, axes=x, values=[lower, upper, y],
        rows_size=rows_size, cols_size=cols_size, padding=padding, title=title, x_axis_names=x_axis_name,
        y_axis_names=y_axis_name, echart_options=common_options, series_options=series_options,
        option_modifications=option_modifications, variant=variant
    )


async def segmented_line_chart(
    self: 'PlotLayer', order: int, x: str, y: Union[str, list[str]],
    data: Optional[Union[str, pd.DataFrame, list[dict]]],
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    marking_lines: Optional[list[int]] = None,
    range_colors: Optional[list[str]] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    """ Create a chart with bars and lines. """
    data_mappings_to_tuples = await self._choose_data(order, data)

    if isinstance(y, str):
        y = [y]

    if not marking_lines:
        marking_lines = [0]

    common_options = get_common_echart_options()
    common_options['xAxis'][0]['boundaryGap'] = False
    # Calculate the width of the visual map with the label of the marking line
    # ( label  legend_ini - legend_end spaces for the legend colors)
    common_options['grid']['right'] = (
        max([len(f' {marking_lines[i + 1]}   {marking_lines[i]} - {marking_lines[i + 1]}        ')
            for i in range(0, len(marking_lines)-1)]
            ) * 6
    )
    common_options.update({
        'visualMap': {
            'top': 50,
            'right': 10,
            'itemHeight': 14 if variant != 'minimal' else 0,
            'showLabel': variant != 'minimal',
            'outOfRange': {'color': '#999'},
            'pieces': [{'min': marking_lines[i], 'max': marking_lines[i + 1]}
                       for i in range(len(marking_lines) - 1)] +
                      [{'min': marking_lines[-1]}],
        },
    })
    if range_colors:
        common_options['visualMap']['color'] = range_colors[::-1]
    series_options = get_common_series_options()
    series_options.update({
        'type': 'line',
        'color': ShimokuPalette.CHART_C5.value,
        'markLine': {
            'data': [{'yAxis': val} for val in marking_lines],
            'label': {'show': variant != 'minimal'},
            'itemStyle': {
                'color': ShimokuPalette.CHART_C5.value,
                'opacity': 0.5
            },
        }
    })

    await self._create_trend_chart(
        data_mapping_to_tuples=data_mappings_to_tuples, axes=x, values=y,
        echart_options=common_options, series_options=series_options,
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name,
        title=title, rows_size=rows_size, cols_size=cols_size, padding=padding,
        option_modifications=option_modifications, variant=variant
    )



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.utils import (calculate_percentages_from_list, interpret_color)
from shimoku.exceptions import BentoboxError

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


async def shimoku_gauge_chart(
    self: 'PlotLayer', order: int,
    value: Union[int, float], name: str = "",
    color: Optional[Union[str, int]] = 1, title: Optional[str] = None,
    rows_size: Optional[int] = 1, cols_size: Optional[int] = 3,
    padding: Optional[str] = None,
    is_percentage: bool = False,
    option_modifications: Optional[dict] = None,
):
    """ Create a pie chart """

    color = interpret_color(color)

    value_font_size = 24 * max(min(rows_size, int(cols_size / 2)), 1)

    if self._bentobox_data:
        value_font_size = 35 * max(int(min(rows_size / 10,
                                   (cols_size - self._bentobox_data['bentoboxSizeColumns']) / 4)), 1)

    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['yAxis']
    del common_options['xAxis']
    del common_options['grid']
    del common_options['toolbox']
    common_options['tooltip']['trigger'] = 'item'
    common_options['grid'] = {
        'left': '5%',
        'right': '5%',
        'top': '5%',
        'bottom': '5%',
        'containLabel': True
    }
    series_options.update({
        'type': 'gauge',
        'startAngle': 180,
        'endAngle': 0,
        'radius': '100%',
        'min': 0,
        'max': 100,
        'pointer': {
            'show': False,
        },
        'progress': {
            'show': True,
            'width': 30 if rows_size > 1 else 25,
            'overlap': False,
            'roundCap': False,
            'clip': False,
            'itemStyle': {
                'borderWidth': 0,
                'borderColor': color,
                'color': color,
            }
        },
        'splitLine': {
            'show': False,
        },
        'axisLine':
            {'lineStyle':
                 {'width': 30 if rows_size > 1 else 25}
             },
        'axisTick': {
            'show': False
        },
        'axisLabel': {
            'show': False,
        },
        'title': {
            'fontSize': 16,
            'fontFamily': 'Rubik',
            'offsetCenter': ['0', '30'],
        },
        'detail': {
            'fontSize': value_font_size,
            'fontFamily': 'Rubik',
            'font': 'inherit',
            'color': '#202A36',
            'borderColor': 'auto',
            'borderWidth': 0,
            'formatter': ('-' if value < 0 else '') + '{value}' + ('%' if is_percentage else ''),
            'valueAnimation': True,
            'offsetCenter': ['0', '-10']
        }
    })
    common_options['series'] = [series_options]

    await self._create_echart(
        fields=[{'name': 'name', 'value': 'value'}], options=common_options, option_modifications=option_modifications,
        data_mapping_to_tuples=await self._choose_data(order, DataFrame([{'name': name, 'value': abs(value)}])),
        order=order, title=title, rows_size=rows_size, cols_size=cols_size, padding=padding)


def shimoku_gauges_group(
        self: 'PlotLayer', gauges_data: Union[DataFrame, list[dict]], order: int,
        rows_size: Optional[int] = None, cols_size: Optional[int] = 12,
        gauges_padding: Optional[str] = '3, 1, 1, 1',
        gauges_rows_size: Optional[int] = 9, gauges_cols_size: Optional[int] = 4,
        calculate_percentages: Optional[bool] = False
):
    if isinstance(gauges_data, DataFrame):
        gauges_data = gauges_data.to_dict(orient="records")

    if calculate_percentages:
        percentages = calculate_percentages_from_list([gauge['value'] for gauge in gauges_data], 0)
        for i in range(len(percentages)):
            gauges_data[i]['value'] = percentages[i]

    if self._bentobox_data:
        log_error(logger, 'The gauges group uses a bentobox, so it cannot be used inside another bentobox. Please'
                          ' pop out of the current bentobox before using this function.', BentoboxError)

    self.set_bentobox(cols_size, rows_size)
    for gauge in gauges_data:
        order += 1
        self.shimoku_gauge(
            value=gauge['value'], order=order, name=gauge.get('name'),
            padding=gauge['padding'] if gauge.get('padding') else gauges_padding,
            color=gauge['color'] if gauge.get('color') else 1,
            rows_size=gauge['rows_size'] if gauge.get('rows_size') else gauges_rows_size,
            cols_size=gauge['cols_size'] if gauge.get('cols_size') else gauges_cols_size,
            is_percentage=gauge['is_percentage'] if gauge.get('is_percentage') else True,
        )

    self.pop_out_of_bentobox()
    return order + 1
